@@ -23,8 +23,8 @@ import matplotlib.image as mgimg
 import numpy as np
 import scipy.ndimage.measurements as snm
 from PIL import Image
-from keras.models import load_model
-from keras.preprocessing.image import ImageDataGenerator
+from tensorflow.keras.models import load_model
+from tensorflow.keras.preprocessing.image import ImageDataGenerator
 from matplotlib import animation
 from matplotlib import pyplot as plt
 from scipy.stats import trim_mean
@@ -61,16 +61,6 @@ def timer(func):
     return wrapper_timer
 
 
-def save_coordinates(obj, filename):
-    """"""
-    # TODO: save_coordinates
-
-
-def load_coordinates(filename):
-    """"""
-    # TODO: load_coordinates
-
-
 def get_random_cmap(num, seed=1):
     """
     Generate a random cmap
@@ -121,7 +111,7 @@ def get_tracking_path(adjacent, ensemble, folder_path):
 
 def read_image_ts(vol, path, name, z_range, print_=False):
     """
-    Load a single volume of the 3D+T sub_images.
+    Read a 3D image at time vol
 
     Parameters
     ----------
@@ -139,7 +129,7 @@ def read_image_ts(vol, path, name, z_range, print_=False):
     Returns
     -------
     img_array : numpy.ndarray
-        An array of the image with shape (x, y, z)
+        An array of the image with shape (row, column, layer)
     """
     image_raw = []
     for z in range(z_range[0], z_range[1]):
@@ -152,14 +142,14 @@ def read_image_ts(vol, path, name, z_range, print_=False):
 
 def save_img3(z_siz, img, path):
     """
-    Save a 3D image as 2D image sequence
+    Save a 3D image (at t=1) as 2D image sequence
 
     Parameters
     ----------
     z_siz : int
         The layer number of the 3D image
     img : numpy.ndarray
-        The 3D image to be saved
+        The 3D image to be saved. Shape: (row, column, layer)
     path : str
         The path of the image files to be saved.
         It should use formatted string to indicate volume number and then layer number, e.g. "xxx_t%04d_z%04i.tif"
@@ -263,8 +253,8 @@ class Draw:
                       cmap=get_random_cmap(num=self.cell_num_t0))
         axm[1].set_title(f"Manual _segment at vol 1", fontdict=TITLE_STYLE)
 
-    def _get_tracking_anim(self, ax, r_coordinates_predicted_pre, r_coordinates_segmented_post,
-                           r_coordinates_predicted_post, layercoord, draw_point=True):
+    def _draw_transformation(self, ax, r_coordinates_predicted_pre, r_coordinates_segmented_post,
+                             r_coordinates_predicted_post, layercoord, draw_point=True):
         """
         Draw each iteration of the tracking by FFN + PR-GLS
         """
@@ -300,7 +290,7 @@ class Draw:
         """
         Draw the accurate correction of cell positions after FFN + PR-GLS transformation
         """
-        _ = self._get_tracking_anim(
+        _ = self._draw_transformation(
             [ax[0], ax[1]],
             self._transform_real_to_layer(r_coor_predicted),
             self._transform_real_to_layer(self.segresult.r_coordinates_segment),
@@ -346,8 +336,8 @@ class Draw:
             ax_i.cla()
         plt.suptitle(f"Tracking results at vol {target_volume}", size=16)
 
-        _ = self._get_tracking_anim([ax[0], ax[1]], self.history.r_tracked_coordinates[target_volume - 2],
-                                    self.segresult.r_coordinates_segment, r_coor_predicted_mean, layercoord=False)
+        _ = self._draw_transformation([ax[0], ax[1]], self.history.r_tracked_coordinates[target_volume - 2],
+                                      self.segresult.r_coordinates_segment, r_coor_predicted_mean, layercoord=False)
         self._draw_correction([ax[2], ax[3]], r_coor_predicted_mean, i_disp_from_vol1_updated)
         self._draw_after_matching(ax[4], ax[5], target_volume, legend=False)
         self._set_layout_anim()
@@ -755,12 +745,12 @@ class History:
 
     Attributes
     ----------
-    r_displacements :  list
-        The displacements of each cell in each volume relative to their positions in volume 1
-    r_segmented_coordinates : list
-        The positions of each cell segmented by 3D U-Net + watershed in each volume
-    r_tracked_coordinates : list
-        The positions of each cell tracked by FFN + PR-GLS + corrections in each volume
+    r_displacements :  list of numpy.ndarray
+        The displacements of each cell in each volume relative to their positions in volume 1. Shape: (cell_num, 3)
+    r_segmented_coordinates : list of numpy.ndarray
+        The positions of each cell segmented by 3D U-Net + watershed in each volume. Shape: (cell_num, 3)
+    r_tracked_coordinates : list of numpy.ndarray
+        The positions of each cell tracked by FFN + PR-GLS + corrections in each volume. Shape: (cell_num, 3)
     anim : list
         The images of tracking process in each volume (from volume 2)
     """
@@ -818,14 +808,14 @@ class Tracker(Segmentation, Draw):
         For padding the images before applying U-Net and for shrinking the cell prediction in the center part
         of the U-Net output. Each value should be < (x, y, or z size of the U-Net input // 2), Default: (24, 24, 2)
     miss_frame : list, optional
-        A list of volumes which includes serious problem in the raw images and thus will be skipped for tracking.
+        A list of volumes (int) which includes serious problem in the raw images and thus will be skipped for tracking.
         Default: None
     cell_num_t0 : int
         The detected cell numbers in the manually corrected segmentation in volume 1
     Z_RANGE_INTERP : Range object
         The sequence of the indexes along z axis in the interpolated image corresponding to the layers in the raw image
     region_list : list
-        List of the sub-images [array_cell1, array_cell2, ...] with binary values
+        List of the 3D sub-images [array_cell1, array_cell2, ...] with binary values
         (1: this cell; 0: background or other cells)
     region_width : list
         List of the width of [[x_width, y_width, z_width]_cell1, ...] each sub-image in x, y, and z axis
@@ -836,9 +826,10 @@ class Tracker(Segmentation, Draw):
     pad_z : int
         The values for padding a zero array with the raw image size
     label_padding : numpy.ndarray
-        The zero array padded with pad_x, pad_y, pad_z, to be used in accurate correction during tracking
+        A 3D array with zero values, used in accurate correction during tracking.
+        Shape: (row + 2 * pad_x, column + 2 * pad_y, layer + 2 * pad_z)
     segmentation_manual_relabels : numpy.ndarray
-        The relabeled manual segmentation
+        The relabeled manual segmentation. Shape: (row, column, layer)
     segresult : SegResults object
         The results of the segmentation in current volume
     unet_model : keras.Model
@@ -990,7 +981,7 @@ class Tracker(Segmentation, Draw):
         for step in range(1, iteration + 1):
             self.unet_model.fit_generator(self.train_generator, validation_data=self.valid_data, epochs=1,
                                           steps_per_epoch=60)
-            loss = self.unet_model.history.history["val_loss"]
+            loss = self.unet_model.history.history["val_loss"][-1]
             if loss < min(self.val_losses):
                 print("val_loss updated from ", min(self.val_losses), " to ", loss)
                 self.unet_model.save_weights(os.path.join(self.paths.unet_weights, weights_name + f"step{step}.h5"))
@@ -1130,11 +1121,12 @@ class Tracker(Segmentation, Draw):
     def match(self, target_volume, method="min_size"):
         """
         Match cells in volume 1 with the target_volume
+
         Parameters
         ----------
         target_volume : int
             The target volume to be matched
-        method :
+        method : str
             The method to segment cells.
 
         Returns
@@ -1199,7 +1191,7 @@ class Tracker(Segmentation, Draw):
             for i in range(len(C_t)):
                 r_coordinates_predicted, r_coordinates_predicted_pre = self._predict_one_rep(
                     r_coordinates_predicted, coor_intermediate_list[i], BETA_t[i], C_t[i])
-                plt_obj = self._get_tracking_anim(
+                plt_obj = self._draw_transformation(
                     ax, r_coordinates_predicted_pre, self.segresult.r_coordinates_segment,
                     r_coordinates_predicted, layercoord=False)
                 plt_objs.append(plt_obj)
@@ -1214,15 +1206,24 @@ class Tracker(Segmentation, Draw):
 
     def _fit_ffn_prgls(self, rep, r_coordinates_segment_pre):
         """
-        Appliy FFN + PR-GLS from t1 to t2 (multiple times) to get transformation
-            parameters to predict cell coordinates
-        Input:
-            rep: the number of repetitions of (FFN + max_iteration times of PR-GLS)
-        Return:
-            C_t: list of C in each repetition (to predict the transformed coordinates)
-            BETA_t: list of the parameter beta used in each repetition (to predict coordinates)
-            coor_intermediate_list: list of the pre-transformed coordinates of automatically
-                segmented cells in each repetition (to predict coordinates)
+        Appliy FFN + PR-GLS from t1 to t2 (multiple times) to get transformation parameters to predict cell coordinates
+
+        Parameters
+        ----------
+        rep : int
+            The number of repetitions of (FFN + max_iteration times of PR-GLS)
+        r_coordinates_segment_pre : numpy.ndarray
+            Coordinates of cells in previous volume. Shape: (cell_num, 3)
+
+        Returns
+        -------
+        C_t : list
+            List of C in each repetition (to predict the transformed coordinates)
+        BETA_t : list
+            List of the parameter beta used in each repetition (to predict coordinates)
+        coor_intermediate_list : list
+            List of the pre-transformed coordinates of automatically segmented cells in each repetition
+            (to predict coordinates)
         """
         corr_intermediate = r_coordinates_segment_pre.copy()
         C_t = []
@@ -1345,7 +1346,7 @@ class Tracker(Segmentation, Draw):
         mask : numpy.ndarray
             The new image with the
         """
-        label_moved = self.label_padding.copy() * 0
+        label_moved = self.label_padding.copy()
         mask = label_moved.copy()
         for label in range(0, len(self.region_list)):
             new_x_min = self.region_xyz_min[label][0] + vectors3d[label, 0] + self.pad_x
@@ -1354,7 +1355,7 @@ class Tracker(Segmentation, Draw):
             subregion_previous = label_moved[new_x_min:new_x_min + self.region_width[label][0],
                                  new_y_min:new_y_min + self.region_width[label][1],
                                  new_z_min:new_z_min + self.region_width[label][2]]
-            if len(subregion_previous.flatten()) == 0:
+            if subregion_previous.shape != self.region_list[label].shape:
                 continue
             subregion_new = subregion_previous * (1 - self.region_list[label]) + \
                             self.region_list[label] * (label + 1)
@@ -1517,3 +1518,18 @@ class Tracker(Segmentation, Draw):
         self.history.r_tracked_coordinates.append(self.r_coordinates_tracked_t0 + r_disp_from_vol1_updated)
 
         return None
+
+    def save_coordinates(self):
+        """Save 3D coordinates in a csv file under the track_information folder
+
+        Notes
+        -----
+        x,y are coordinates with pixel unit, while z is the interpolated coordinate with the same unit as x and y
+        """
+        coord = np.asarray(self.history.r_tracked_coordinates)
+        t, cell, pos = coord.shape
+        coord_table = np.column_stack(
+            (np.repeat(np.arange(1, t+1), cell), np.tile(np.arange(1, cell+1), t), coord.reshape(t * cell, pos)))
+        np.savetxt(os.path.join(self.paths.track_information, "tracked_coordinates.csv"), coord_table, delimiter=',',
+                   header="cell,t,x(row),y(column),z(interpolated)", comments="")
+        print("Cell coordinates were stored in ./track_information/tracked_coordinates.csv")
